#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-4-30 上午9:36
@File    : __init__.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import os

import pandas as pd

from ibats_common.backend.mess import get_folder_path

OHLCAV_COL_NAME_LIST = ["open", "high", "low", "close", "amount", "volume"]


def load_data(file_name, encoding=None, folder_path=None, index_col=None, range_from=None,
              range_to=None) -> pd.DataFrame:
    if folder_path is None:
        folder_path = os.path.join(get_folder_path('example', create_if_not_found=False), 'data')
        if folder_path is None:
            raise ValueError(
                "folder_path is None, or you have to had a ./example/data folder on current or parent folder")
    file_path = os.path.join(folder_path, file_name)
    df = pd.read_csv(file_path, encoding=encoding, index_col=index_col,
                     parse_dates=[index_col] if index_col is not None else False)
    # df.index = pd.DatetimeIndex(df.index)
    # 获取指定日期区间的数据
    if range_from is None:
        is_in_range = None
    else:
        is_in_range = df.index >= pd.to_datetime(range_from)
    if range_to is not None:
        if is_in_range is None:
            is_in_range = df.index <= pd.to_datetime(range_to)
        else:
            is_in_range &= df.index <= pd.to_datetime(range_to)
    if is_in_range is not None:
        df = df[is_in_range]

    return df


def get_trade_date_series(folder_path=None):
    df = load_data('trade_date.csv', folder_path=folder_path).astype('datetime64[ns]')
    date_s = df.iloc[:, 0]
    return date_s


def get_delivery_date_series(instrument_type, folder_path=None):
    df = load_data(
        'future_info.csv', folder_path=folder_path
    ).set_index(
        'symbol'
    ).filter(
        regex='^' + instrument_type + r'(?=\d+$)', axis=0
    )
    # re_pattern_instrument_header = re.compile(r'[A-Za-z]+(?=\d+$)')
    date_s = df["delist_date"].astype('datetime64[ns]').sort_values()
    # ret_list = [str_2_date(_) for _ in load_data('trade_date.csv').T.to_numpy()[0]]
    return date_s


def _test_load_data():
    from ibats_utils.mess import is_windows_os, str_2_date
    DATA_FOLDER_PATH = r'D:\WSPych\IBATSCommon\ibats_common\example\data' \
        if is_windows_os() else r'/home/mg/github/IBATS_Common/ibats_common/example/data'
    df = load_data('RB.csv',
                   folder_path=DATA_FOLDER_PATH,
                   index_col='trade_date', range_from='2013-1-1', range_to='2015-12-31'
                   )
    assert min(df.index) > pd.to_datetime('2013-1-1')
    assert max(df.index) == pd.to_datetime('2015-12-31')


if __name__ == "__main__":
    _test_load_data()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-6-21 上午9:59
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
module_version = 'v3'


if __name__ == "__main__":
    pass



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-5-30 上午8:40
@File    : rnn3_stg.py
@contact : mmmaaaggg@163.com
@desc    : CNN 三分类预测
pip3 install tensorflow sklearn tflearn
2019-06-06
1）  模型及日志目录结构调整如下：
    tf_saves_2019-06-05_16_21_39
      *   model_tfls
      *       *   2012-12-31
      *       *       *   checkpoint
      *       *       *   model_-54_51.tfl.data-00000-of-00001
      *       *       *   model_-54_51.tfl.index
      *       *       *   model_-54_51.tfl.meta
      *       *   2013-02-28
      *       *       *   checkpoint
      *       *       *   model_-54_51.tfl.data-00000-of-00001
      *       *       *   model_-54_51.tfl.index
      *       *       *   model_-54_51.tfl.meta
      *   tensorboard_logs
      *       *   2012-12-31_496[1]_20190605_184316
      *       *       *   events.out.tfevents.1559731396.mg-ubuntu64        datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
    self.base_folder_path = folder_path = os.path.join(get_report_folder_path(self.stg_run_id), f'tf_saves_{datetime_str}')

      *       *   2013-02-28_496[1]_20190605_184716
      *       *       *   events.out.tfevents.1559731396.mg-ubuntu64

2）  调整restore功能
3）  增加模型对未来数据预测成功率走势图
2019-06-12
对原始数据的 OHLCA 乘以一个因子以扩充样本数据，相应的调整 get_factor 方法及相关代码
2019-07-01
继承自 AIStgBase 类
"""
import os
import random
from collections import defaultdict
from datetime import datetime

import numpy as np
import tflearn
from tflearn import conv_2d, max_pool_2d, local_response_normalization, fully_connected, dropout

from ibats_common.backend.mess import get_report_folder_path
from ibats_common.backend.factor import get_factor
from ibats_common.backend.label import calc_label3
from ibats_common.common import BacktestTradeMode, ContextKey, Direction, CalcMode
from ibats_common.example import AIStgBase
from ibats_common.example.data import get_trade_date_series, get_delivery_date_series
from ibats_common.strategy_handler import strategy_handler_factory
from ibats_local_trader.agent.md_agent import *
from ibats_local_trader.agent.td_agent import *

logger = logging.getLogger(__name__)


class AIStg(AIStgBase):

    def __init__(self, instrument_type, unit=1):
        super().__init__(instrument_type, unit=1)
        self.unit = unit
        # 模型相关参数
        self.input_size = 38
        self.batch_size = 128
        self.n_step = 60
        self.output_size = 3
        self.n_hidden_units = 10
        self.lr = 0.006
        # 模型训练，及数据集相关参数
        self._model = None
        self._session = None
        self.train_validation_rate = 0.8
        self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = None, None, None, None
        self.label_func_max_rr = 0.0051
        self.label_func_min_rr = -0.0054
        self.max_future = 3
        self.predict_test_random_state = None
        self.n_epoch = 48
        self.retrain_period = 60
        self.validation_accuracy_base_line = 0.55  # 0.6    # 如果为 None，则不进行 validation 成功率检查
        self.over_fitting_train_acc = 0.9  # 过拟合训练集成功率，如果为None则不进行判断
        # 其他辅助信息
        self.trade_date_series = get_trade_date_series()
        self.delivery_date_series = get_delivery_date_series(instrument_type)
        self.tensorboard_verbose = 3
        # 模型保存路径相关参数
        # 是否使用现有模型进行操作，如果是记得修改以下下方的路径
        # enable_load_model_if_exist 将会在调用 self.load_model_if_exist 时进行检查
        # 如果该字段为 False，调用 load_model_if_exist 时依然可以传入参数的方式加载已有模型
        # 该字段与 self.load_model_if_exist 函数的 enable_load_model_if_exist参数是 “or” 的关系
        self.enable_load_model_if_exist = False
        if self.enable_load_model_if_exist:
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), f'tf_saves_2019-06-16_09_33_30')
        else:
            datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), f'tf_saves_{datetime_str}')

        self.model_folder_path = model_folder_path = os.path.join(folder_path, 'model_tfls')
        if not os.path.exists(model_folder_path):
            os.makedirs(model_folder_path, exist_ok=True)
        self.tensorboard_dir = tensorboard_dir = os.path.join(folder_path, f'tensorboard_logs')
        if not os.path.exists(tensorboard_dir):
            os.makedirs(tensorboard_dir)
        self.trade_date_last_train = None
        self.trade_date_acc_list = defaultdict(lambda: [0.0, 0.0])
        self.do_nothing_on_min_bar = False  # 仅供调试使用
        # 用于记录 open,high,low,close,amount 的 column 位置
        self.ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]

    def get_x_y(self, factor_df):
        factors, trade_date_index = self.get_factor_array(factor_df)
        price_arr = factors[:, 0]
        self.input_size = factors.shape[1]
        # ys_all = self.calc_y_against_future_data(price_arr, -0.01, 0.01)
        # ys_all = self.calc_y_against_future_data(price_arr, -self.classify_wave_rate, self.classify_wave_rate)
        ys_all = calc_label3(price_arr, self.label_func_min_rr, self.label_func_max_rr,
                             max_future=self.max_future, one_hot=True)
        range_from = self.n_step
        range_to = factors.shape[0]

        xs = np.zeros((range_to - range_from, self.n_step, self.input_size, 1))
        for num, index in enumerate(range(range_from, range_to)):
            xs[num, :, :, :] = factors[(index - self.n_step):index, :, np.newaxis]

        ys = ys_all[range_from:range_to, :]

        return xs, ys, trade_date_index[range_from:range_to]

    def get_batch_xs(self, factors: np.ndarray, index=None):
        """
        取 batch_xs
        :param factors:
        :param index: 样本起始坐标，如果为None，则默认取尾部一组样本
        :return:
        """
        if index is None:
            index = factors.shape[0] - 1

        batch_xs = np.zeros((1, self.n_step, self.input_size, 1))
        batch_xs[0, :, :, :] = factors[(index - self.n_step + 1):(index + 1), :, np.newaxis]

        return batch_xs

    def get_batch_by_random(self, factors: np.ndarray, labels: np.ndarray):
        """
        够在一系列输入输出数据集
        xs： 两条同频率，不同位移的sin曲线
        ys_value： 目标是一条cos曲线
        ys: ys_value 未来涨跌标识
        i_s：X 序列
        """
        xs = np.zeros((self.batch_size, self.n_step, self.input_size))
        ys = np.zeros((self.batch_size, self.output_size))
        # available_batch_size, num = 0, 0
        samples_index = random.sample(range(self.n_step - 1, factors.shape[0] - 1), self.batch_size)
        examples_index_list = []
        for available_batch_size, index in enumerate(samples_index):
            tmp = factors[(index - self.n_step + 1):(index + 1), :]
            if tmp.shape[0] < self.n_step:
                break
            xs[available_batch_size, :, :] = tmp
            ys[available_batch_size, :] = labels[index, :]
            examples_index_list.append(index)
            if available_batch_size + 1 >= self.batch_size:
                available_batch_size += 1
                break

        # returned xs, ys_value and shape (batch, step, input)
        return xs, ys, examples_index_list

    def _build_model(self) -> tflearn.models.DNN:
        # Network building
        net = tflearn.input_data([None, self.n_step, self.input_size, 1])  # input_size=39
        net = tflearn.layers.normalization.batch_normalization(net)
        # layer1
        net = conv_2d(net, 32, 3, activation='relu', name='Conv2D_1')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)
        # layer2
        net = conv_2d(net, 64, 3, activation='relu', name='Conv2D_2')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        # layer3
        net = conv_2d(net, 128, 3, activation='relu', name='Conv2D_3')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        # layer4
        net = conv_2d(net, 256, 3, activation='relu', name='Conv2D_4')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        net = fully_connected(net, 1024, activation='tanh', name='fc1')
        net = dropout(net, 0.4, name='Dropout1')
        net = fully_connected(net, 1024, activation='tanh', name='fc2')
        net = dropout(net, 0.4, name='Dropout2')
        net = fully_connected(net, self.output_size, activation='softmax')

        net = tflearn.regression(net, optimizer='momentum',
                                 learning_rate=0.001,
                                 loss='categorical_crossentropy')

        # Training
        _model = tflearn.DNN(net, tensorboard_verbose=self.tensorboard_verbose,
                             # checkpoint_path=self.checkpoint_path,
                             tensorboard_dir=self.tensorboard_dir)
        return _model

    def on_min1(self, md_df, context):
        if self.do_nothing_on_min_bar:  # 仅供调试使用
            return

        # 数据整理
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        indexed_df.index = pd.DatetimeIndex(indexed_df.index)
        # 获取最新交易日
        trade_date = str_2_date(indexed_df.index[-1])
        days_after_last_train = (trade_date - self.trade_date_last_train).days
        if self.retrain_period is not None and 0 < self.retrain_period < days_after_last_train:
            # 重新训练
            self.logger.info('当前日期 %s 距离上一次训练 %s 已经过去 %d 天，重新训练',
                             trade_date, self.trade_date_last_train, days_after_last_train)
            factor_df = self.load_train_test(indexed_df, rebuild_model=True,
                                             enable_load_model=self.enable_load_model_if_exist)
        else:
            factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                   trade_date_series=self.trade_date_series,
                                   delivery_date_series=self.delivery_date_series)

        # 预测
        pred_mark = self.predict_latest(factor_df)
        is_holding, is_buy, is_sell = pred_mark == 0, pred_mark == 1, pred_mark == 2
        # self.logger.info('%s is_buy=%s, is_sell=%s', trade_date, str(is_buy), str(is_sell))
        close = md_df['close'].iloc[-1]
        instrument_id = context[ContextKey.instrument_id_list][0]
        if is_buy:  # is_buy
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_target_position = False
            if no_target_position:
                self.open_long(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_sell:  # is_sell
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_holding:
            self.logger.debug("%s %s * * %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)


def _test_use(is_plot):
    from ibats_common.backend.mess import get_folder_path
    import os
    instrument_type = 'RB'
    # 参数设置
    run_mode = RunMode.Backtest_FixPercent
    calc_mode = CalcMode.Normal
    strategy_params = {'instrument_type': instrument_type, 'unit': 1}
    md_agent_params_list = [{
        'md_period': PeriodType.Min1,
        'instrument_id_list': [instrument_type],
        'datetime_key': 'trade_date',
        'init_md_date_from': '1995-1-1',  # 行情初始化加载历史数据，供策略分析预加载使用
        'init_md_date_to': '2013-1-1',
        # 'C:\GitHub\IBATS_Common\ibats_common\example\ru_price2.csv'
        'file_path': os.path.abspath(os.path.join(
            get_folder_path('example', create_if_not_found=False), 'data', f'{instrument_type}.csv')),
        'symbol_key': 'instrument_type',
    }]
    if run_mode == RunMode.Realtime:
        trade_agent_params = {
        }
        strategy_handler_param = {
        }
    elif run_mode == RunMode.Backtest:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            'init_cash': 10000,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2013-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    else:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2013-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    # 初始化策略处理器
    stghandler = strategy_handler_factory(
        stg_class=AIStg,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        exchange_name=ExchangeName.LocalFile,
        run_mode=run_mode,
        trade_agent_params=trade_agent_params,
        strategy_handler_param=strategy_handler_param,
    )
    stghandler.start()
    time.sleep(10)
    stghandler.keep_running = False
    stghandler.join()
    stg_run_id = stghandler.stg_run_id
    logging.info("执行结束 stg_run_id = %d", stg_run_id)

    if is_plot:
        from ibats_common.analysis.summary import summary_stg_2_docx
        from ibats_utils.mess import open_file_with_system_app
        file_path = summary_stg_2_docx(stg_run_id, enable_clean_cache=False)
        if file_path is not None:
            open_file_with_system_app(file_path)

    return stg_run_id


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    is_plot = True
    _test_use(is_plot)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/6/28 17:09
@File    : market.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import numpy as np
import pandas as pd


class QuotesMarket(object):
    def __init__(self, md_df: pd.DataFrame, data_factors, init_cash=2e5, fee_rate=3e-3,
                 state_with_flag=False, reward_with_fee0=False):
        self.data_close = md_df['close']
        self.data_open = md_df['open']
        self.data_observation = data_factors
        self.action_space = ['close', 'long', 'short', 'keep']
        self.fee_rate = fee_rate  # 千三手续费
        self.fee_curr_step = 0
        self.fee_tot = 0
        self.max_step_count = self.data_observation.shape[0] - 1
        self.init_cash = init_cash
        # reset use
        self.step_counter = 0
        self.cash = self.init_cash
        self.position = 0
        self.total_value = self.cash + self.position
        self.total_value_fee0 = self.cash + self.position
        self.flags = 0
        self.state_with_flag = state_with_flag
        self.reward_with_fee0 = reward_with_fee0
        self.action_count = 0

    def reset(self):
        self.step_counter = 0
        self.cash = self.init_cash
        self.position = 0
        self.total_value = self.cash + self.position
        self.total_value_fee0 = self.cash + self.position
        self.flags = 0
        self.fee_curr_step = 0
        self.fee_tot = 0
        self.action_count = 0
        if self.state_with_flag:
            return self.data_observation[0], np.array([self.flags])
        else:
            return self.data_observation[0]

    def latest_state(self):
        if self.state_with_flag:
            return self.data_observation[-1], np.array([self.flags])
        else:
            return self.data_observation[-1]

    def get_action_space(self):
        return self.action_space

    def long(self):
        self.flags = 1
        quotes = self.data_open[self.step_counter] * 10
        self.cash -= quotes * (1 + self.fee_rate)
        self.position = quotes
        self.fee_curr_step += quotes * self.fee_rate
        self.action_count += 1

    def short(self):
        self.flags = -1
        quotes = self.data_open[self.step_counter] * 10
        self.cash += quotes * (1 - self.fee_rate)
        self.position = - quotes
        self.fee_curr_step += quotes * self.fee_rate
        self.action_count += 1

    def keep(self):
        quotes = self.data_open[self.step_counter] * 10
        self.position = quotes * self.flags

    def close_long(self):
        self.flags = 0
        quotes = self.data_open[self.step_counter] * 10
        self.cash += quotes * (1 - self.fee_rate)
        self.position = 0
        self.fee_curr_step += quotes * self.fee_rate
        self.action_count += 1

    def close_short(self):
        self.flags = 0
        quotes = self.data_open[self.step_counter] * 10
        self.cash -= quotes * (1 + self.fee_rate)
        self.position = 0
        self.fee_curr_step += quotes * self.fee_rate
        self.action_count += 1

    def step_op(self, action):
        self.fee_curr_step = 0
        if action == 'long':
            if self.flags == 0:
                self.long()
            elif self.flags == -1:
                self.close_short()
                self.long()
            else:
                self.keep()

        elif action == 'close':
            if self.flags == 1:
                self.close_long()
            elif self.flags == -1:
                self.close_short()
            else:
                pass

        elif action == 'short':
            if self.flags == 0:
                self.short()
            elif self.flags == 1:
                self.close_long()
                self.short()
            else:
                self.keep()
        elif action == 'keep':
            self.keep()
        else:
            raise ValueError("action should be elements of ['long', 'short', 'close']")

        # 计算费用
        self.fee_tot += self.fee_curr_step
        self.total_value_fee0 = self.total_value + self.fee_tot

        # 计算价值
        price = self.data_close[self.step_counter]
        position = price * 10 * self.flags
        reward = self.cash + position - self.total_value
        self.step_counter += 1
        self.total_value = position + self.cash
        next_observation = self.data_observation[self.step_counter]

        done = False
        if self.total_value < price:
            done = True
        if self.step_counter >= self.max_step_count:
            done = True

        ret_state = (next_observation, np.array([self.flags])) if self.state_with_flag else next_observation
        ret_reward = (reward / price, (reward + self.fee_curr_step) / price) if self.reward_with_fee0 else (
                reward / price)

        return ret_state, ret_reward, done

    def step(self, action):
        if 0 <= action <= 3:
            return self.step_op(self.action_space[action])
        else:
            raise ValueError("action should be one of [0,1,2]")


def _test_quote_market():
    n_step = 60
    ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]
    from ibats_common.example.data import load_data
    md_df = load_data('RB.csv').set_index('trade_date')[ohlcav_col_name_list]
    md_df.index = pd.DatetimeIndex(md_df.index)
    from ibats_common.backend.factor import get_factor, transfer_2_batch
    factors_df = get_factor(md_df, dropna=True)
    df_index, df_columns, data_arr_batch = transfer_2_batch(factors_df, n_step=n_step)
    md_df = md_df.loc[df_index, :]
    # 建立 QuotesMarket
    qm = QuotesMarket(md_df=md_df[['close', 'open']], data_factors=data_arr_batch, state_with_flag=True)
    next_observation = qm.reset()
    assert len(next_observation) == 2
    assert next_observation[0].shape[0] == n_step
    assert next_observation[1] == 0
    next_observation, reward, done = qm.step(1)
    assert len(next_observation) == 2
    assert next_observation[1] == 1
    assert not done
    next_observation, reward, done = qm.step(0)
    assert next_observation[1] == 0
    assert reward != 0
    next_observation, reward, done = qm.step(0)
    assert next_observation[1] == 0
    assert reward == 0
    next_observation, reward, done = qm.step(3)
    assert next_observation[1] == 0
    assert reward == 0
    next_observation, reward, done = qm.step(2)
    assert next_observation[1] == -1
    assert not done
    next_observation, reward, done = qm.step(3)
    assert next_observation[1] == -1
    assert reward != 0
    try:
        qm.step(4)
    except ValueError:
        print('is ok for not supporting action>3')


if __name__ == "__main__":
    _test_quote_market()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-8-8 下午12:28
@File    : utils.py
@contact : mmmaaaggg@163.com
@desc    : use_cup_only 需要优先于一切 import tensorflow 的语句被调用才能生效
"""


def show_device():
    from tensorflow.python.client import device_lib
    return device_lib.list_local_devices()


def _test_show_device():
    devices = show_device()
    print(type(devices), len(devices))
    for num, d in enumerate(devices):
        print(num, '\n', d)


def use_cup_only():
    import os
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # see issue #152
    # os.environ["CUDA_VISIBLE_DEVICES"] = ""
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    import keras.backend.tensorflow_backend as ktf

    import tensorflow as tf
    ktf.set_session(tf.Session(config=tf.ConfigProto(device_count={'gpu': 0})))


if __name__ == '__main__':
    _test_show_device()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/11/7 10:42
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import logging
import os
import re
from collections import defaultdict
from datetime import datetime, timedelta

import ffn
import numpy as np
import pandas as pd
import tensorflow as tf
import tflearn
from ibats_utils.mess import copy_module_file_to, date_2_str, str_2_date, get_last, copy_file_to
from sklearn.model_selection import train_test_split

from ibats_common.backend.mess import get_report_folder_path
from ibats_common.analysis.plot import show_dl_accuracy
from ibats_common.analysis.summary import summary_release_2_docx
from ibats_common.backend.factor import get_factor
from ibats_common.backend.label import calc_label3
from ibats_common.common import ContextKey, Direction
from ibats_common.example.data import get_trade_date_series, get_delivery_date_series
from ibats_common.strategy import StgBase

logger = logging.getLogger(__name__)
logger.debug('import %s', ffn)


class AIStgBase(StgBase):

    def __init__(self, instrument_type, unit=1):
        super().__init__()
        self.unit = unit
        # 模型相关参数
        self.input_size = 38
        self.batch_size = 512
        self.n_step = 60
        self.output_size = 3
        self.n_hidden_units = 72
        self.lr = 0.006
        # 模型训练，及数据集相关参数
        self._model = None
        self._session = None
        self.train_validation_rate = 0.8
        self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = None, None, None, None
        self.label_func_max_rr = 0.0051
        self.label_func_min_rr = -0.0054
        self.max_future = 3
        self.predict_test_random_state = None
        self.n_epoch = 40
        self.max_loop_4_futher_train = 20
        self.retrain_period = 30  # 60 每隔60天重新训练一次，0 则为不进行重新训练
        self.validation_accuracy_base_line = 0.55  # 0.6    # 如果为 None，则不进行 validation 成功率检查
        self.over_fitting_train_acc = 0.95  # 过拟合训练集成功率，如果为None则不进行判断
        # 其他辅助信息
        self.trade_date_series = get_trade_date_series()
        self.delivery_date_series = get_delivery_date_series(instrument_type)
        self.tensorboard_verbose = 3
        # 模型保存路径相关参数
        # 是否使用现有模型进行操作，如果是记得修改以下下方的路径
        # enable_load_model_if_exist 将会在调用 self.load_model_if_exist 时进行检查
        # 如果该字段为 False，调用 load_model_if_exist 时依然可以传入参数的方式加载已有模型
        # 该字段与 self.load_model_if_exist 函数的 enable_load_model_if_exist参数是 “or” 的关系
        self.enable_load_model_if_exist = False
        if self.enable_load_model_if_exist:
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), 'tf_saves_2019-06-27_16_24_34')
        else:
            datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), f'tf_saves_{datetime_str}')
        # 备份文件
        file_path = copy_module_file_to(self.__class__, folder_path)
        self.logger.debug('文件已经备份到：%s', file_path)
        # 设置模型备份目录
        self.model_folder_path = model_folder_path = os.path.join(folder_path, 'model_tfls')
        if not os.path.exists(model_folder_path):
            os.makedirs(model_folder_path, exist_ok=True)
        # 设置日至备份目录
        self.tensorboard_dir = tensorboard_dir = os.path.join(folder_path, f'tensorboard_logs')
        if not os.path.exists(tensorboard_dir):
            os.makedirs(tensorboard_dir)
        self.trade_date_last_train = None
        self.trade_date_acc_list = defaultdict(lambda: [0.0, 0.0])
        self.do_nothing_on_min_bar = False  # 仅供调试使用
        # 用于记录 open,high,low,close,amount 的 column 位置
        self.ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]

    @property
    def session(self):
        return self.get_session()

    def get_session(self, renew=False, close_last_one_if_renew=True):
        if renew or self._session is None:
            if self.model is None:
                raise ValueError('model 需要先于 session 被创建')
            if close_last_one_if_renew and self._session is not None:
                self._session.close()
            self._session = tf.Session()
            self._session.run(tf.global_variables_initializer())
        return self._session

    @property
    def model(self):
        """
        获取模型对象
        2019-07-05 去除  -> tflearn.models.DNN hint 该语句将会导致动态加载 class 时抛出错误
        :return:
        """
        return self.get_model()

    def get_model(self, rebuild_model=False):
        """
        获取模型对象，默认不重新构造
        2019-07-05 去除  -> tflearn.models.DNN hint 该语句将会导致动态加载 class 时抛出错误
        :param rebuild_model:
        :return:
        """
        if self._model is None or rebuild_model:
            self.logger.info('重新构建模型')
            tf.reset_default_graph()
            self._model = self._build_model()
        return self._model

    def get_factor_array(self, md_df: pd.DataFrame, tail_n=None):
        if tail_n is not None:
            md_df = md_df.tail(tail_n)
        df = md_df[~md_df['close'].isnull()]

        factors = df.fillna(0).to_numpy()
        if self.input_size is None or self.input_size != factors.shape[1]:
            self.input_size = factors.shape[1]
            self.n_hidden_units = self.input_size * 2
            self.logger.info("set input_size: %d", self.input_size)
            self.logger.info("set n_hidden_units: %d", self.n_hidden_units)

        return factors, df.index

    def get_x_y(self, factor_df):
        factors, trade_date_index = self.get_factor_array(factor_df)
        price_arr = factors[:, 0]
        self.input_size = factors.shape[1]
        # ys_all = self.calc_y_against_future_data(price_arr, -0.01, 0.01)
        # ys_all = self.calc_y_against_future_data(price_arr, -self.classify_wave_rate, self.classify_wave_rate)
        ys_all = calc_label3(price_arr, self.label_func_min_rr, self.label_func_max_rr,
                             max_future=self.max_future, one_hot=True)
        range_from = self.n_step
        range_to = factors.shape[0]

        xs = np.zeros((range_to - range_from, self.n_step, self.input_size))
        for num, index in enumerate(range(range_from, range_to)):
            xs[num, :, :] = factors[(index - self.n_step):index, :]

        ys = ys_all[range_from:range_to, :]

        return xs, ys, trade_date_index[range_from:range_to]

    def get_batch_xs(self, factors: np.ndarray, index=None):
        """
        取 batch_xs
        :param factors:
        :param index: 样本起始坐标，如果为None，则默认取尾部一组样本
        :return:
        """
        if index is None:
            index = factors.shape[0] - 1

        batch_xs = np.zeros((1, self.n_step, self.input_size))
        batch_xs[0, :, :] = factors[(index - self.n_step + 1):(index + 1), :]

        return batch_xs

    def _build_model(self):
        # Network building
        # net = tflearn.input_data([None, self.n_step, self.input_size])
        # ...
        # Training
        # _model = tflearn.DNN(net, tensorboard_verbose=self.tensorboard_verbose,
        #                      checkpoint_path=self.checkpoint_path,
        #                      tensorboard_dir=self.tensorboard_dir)
        # return _model
        raise NotImplementedError()

    def train(self, factor_df_dic: dict, predict_test_random_state):
        factor_df = factor_df_dic[1]

        trade_date_from_str, trade_date_to_str = date_2_str(factor_df.index[0]), date_2_str(factor_df.index[-1])
        # xs_train, xs_validation, ys_train, ys_validation = self.separate_train_validation(xs, ys)
        if self.predict_test_random_state is None:
            random_state = predict_test_random_state
        else:
            random_state = self.predict_test_random_state

        # 利用生成数据做训练数据集，只用原始数据中的 validation 部分做验证集
        arr_list, xs_validation, ys_validation = [], None, None
        for adj_factor, factor_df in factor_df_dic.items():
            xs, ys, _ = self.get_x_y(factor_df)
            xs_train_tmp, xs_validation_tmp, ys_train_tmp, ys_validation_tmp = train_test_split(
                xs, ys, test_size=0.2, random_state=random_state)
            arr_list.append([xs_train_tmp, ys_train_tmp])
            # xs_train, xs_validation = xs_train_tmp, xs_validation_tmp
            # ys_train, ys_validation = ys_train_tmp, ys_validation_tmp
            if adj_factor == 1:
                xs_validation, ys_validation = xs_validation_tmp, ys_validation_tmp

        self.xs_train = xs_train = np.vstack([_[0] for _ in arr_list])
        self.ys_train = ys_train = np.vstack([_[1] for _ in arr_list])

        sess = self.get_session(renew=True)
        train_acc, val_acc = 0, 0
        with sess.as_default():
            # with tf.Graph().as_default():
            # self.logger.debug('sess.graph:%s tf.get_default_graph():%s', sess.graph, tf.get_default_graph())
            self.logger.debug('[%d], xs_train %s, ys_train %s, xs_validation %s, ys_validation %s, [%s, %s]',
                              random_state, xs_train.shape, ys_train.shape, xs_validation.shape, ys_validation.shape,
                              trade_date_from_str, trade_date_to_str)
            max_loop = self.max_loop_4_futher_train
            for num in range(max_loop):
                if num == 0:
                    n_epoch = self.n_epoch
                else:
                    n_epoch = self.n_epoch // max_loop

                self.logger.info('[%d]第 %d/%d 轮训练，开始 [%s, %s] n_epoch=%d', random_state, num + 1, max_loop,
                                 trade_date_from_str, trade_date_to_str, n_epoch)
                run_id = f'{trade_date_to_str}_{xs_train.shape[0]}[{predict_test_random_state}]' \
                         f'_{datetime.now().strftime("%Y%m%d_%H%M%S")}'
                tflearn.is_training(True)
                self.model.fit(
                    xs_train, ys_train, validation_set=(xs_validation, ys_validation),
                    show_metric=True, batch_size=self.batch_size, n_epoch=n_epoch,
                    run_id=run_id)
                tflearn.is_training(False)

                result = self.model.evaluate(xs_train, ys_train, batch_size=self.batch_size)
                # self.logger.info("train accuracy: %.2f%%" % (result[0] * 100))
                train_acc = result[0]
                result = self.model.evaluate(xs_validation, ys_validation, batch_size=self.batch_size)
                val_acc = result[0]
                self.logger.info("[%d]第 %d/%d 轮训练，[%s - %s]，训练集准确率(train_acc)：%.2f%%， 样本外准确率(val_acc): %.2f%%",
                                 random_state, num + 1, max_loop, trade_date_from_str, trade_date_to_str,
                                 train_acc * 100, val_acc * 100)
                if self.over_fitting_train_acc is not None and train_acc > self.over_fitting_train_acc:
                    self.logger.warning('[%d]第 %d/%d 轮训练，训练集精度 %.2f%% > %.2f%% 可能存在过拟合 [%s, %s]',
                                        random_state, num + 1, max_loop, train_acc * 100,
                                        self.over_fitting_train_acc * 100,
                                        trade_date_from_str, trade_date_to_str)
                    break
                if self.validation_accuracy_base_line is not None:
                    if result[0] > self.validation_accuracy_base_line:
                        break
                    elif num < max_loop - 1:
                        self.logger.warning('[%d]第 %d/%d 轮训练，[%s - %s]，样本外训练准确率 %.2f%% < %.0f%%，继续训练',
                                            random_state, num + 1, max_loop, trade_date_from_str, trade_date_to_str,
                                            val_acc * 100, self.validation_accuracy_base_line * 100)
                else:
                    break

        self.trade_date_last_train = str_2_date(trade_date_to_str)
        return train_acc, val_acc

    def valid_model_acc(self, factor_df: pd.DataFrame):
        xs, ys, _ = self.get_x_y(factor_df)
        trade_date_from_str, trade_date_to_str = date_2_str(factor_df.index[0]), date_2_str(factor_df.index[-1])
        random_state = self.predict_test_random_state
        xs_train, xs_validation, ys_train, ys_validation = train_test_split(
            xs, ys, test_size=0.2, random_state=random_state)
        self.logger.debug('random_state=%d, xs_train %s, ys_train %s, xs_validation %s, ys_validation %s, [%s, %s]',
                          random_state, xs_train.shape, ys_train.shape, xs_validation.shape, ys_validation.shape,
                          trade_date_from_str, trade_date_to_str)
        sess = self.get_session(renew=True)
        with sess.as_default():
            # with tf.Graph().as_default():
            # self.logger.debug('sess.graph:%s tf.get_default_graph():%s', sess.graph, tf.get_default_graph())
            result = self.model.evaluate(xs_validation, ys_validation, batch_size=self.batch_size)
            val_acc = result[0]
            result = self.model.evaluate(xs_train, ys_train, batch_size=self.batch_size)
            train_acc = result[0]
            self.logger.info("[%s - %s] 训练集准确率: %.2f%%", trade_date_from_str, trade_date_to_str, train_acc * 100)
            self.logger.info("[%s - %s] 验证集准确率: %.2f%%", trade_date_from_str, trade_date_to_str, val_acc * 100)
        return train_acc, val_acc

    def save_model(self, trade_date):
        """
        将模型导出到文件
        :param trade_date:
        :return:
        """
        folder_path = os.path.join(self.model_folder_path, date_2_str(trade_date))
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        file_path = os.path.join(
            folder_path,
            f"model_{int(self.label_func_min_rr * 10000)}_{int(self.label_func_max_rr * 10000)}.tfl")
        self.model.save(file_path)
        self.logger.info("模型训练截止日期： %s 保存到: %s", self.trade_date_last_train, file_path)
        return file_path

    def load_model_if_exist(self, trade_date, enable_load_model_if_exist=False):
        """
        将模型导出到文件
        目录结构：
        tf_saves_2019-06-05_16_21_39
          *   model_tfls
          *       *   2012-12-31
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *       *   2013-02-28
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *   tensorboard_logs
          *       *   2012-12-31_496[1]_20190605_184316
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
          *       *   2013-02-28_496[1]_20190605_184716
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
        :param enable_load_model_if_exist:
        :param trade_date:
        :return:
        """
        if self.enable_load_model_if_exist or enable_load_model_if_exist:
            # 获取小于等于当期交易日的最大的一个交易日对应的文件名
            min_available_date = str_2_date(trade_date) - timedelta(days=self.retrain_period)
            self.logger.debug('尝试加载现有模型，[%s - %s] %d 天', min_available_date, trade_date, self.retrain_period)
            date_file_path_pair_list = [_ for _ in self.get_date_file_path_pair_list() if _[0] >= min_available_date]
            if len(date_file_path_pair_list) > 0:
                # 按日期排序
                date_file_path_pair_list.sort(key=lambda x: x[0])
                # 获取小于等于当期交易日的最大的一个交易日对应的文件名
                # file_path = get_last(date_file_path_pair_list, lambda x: x[0] <= trade_date, lambda x: x[1])
                trade_date = str_2_date(trade_date)
                ret = get_last(date_file_path_pair_list, lambda x: x[0] <= trade_date)
                if ret is not None:
                    key, folder_path, predict_test_random_state = ret
                    if folder_path is not None:
                        model = self.get_model(rebuild_model=True)  # 这句话是必须的，需要实现建立模型才可以加载
                        model.load(folder_path)
                        self.trade_date_last_train = key
                        self.predict_test_random_state = predict_test_random_state
                        self.logger.info("加载模型成功。trade_date_last_train: %s load from path: %s", key, folder_path)
                        return True

        return False

    def predict_test(self, md_df):
        self.logger.info('开始预测')
        # sess = self.session
        if self.xs_train is None:
            xs, ys, _ = self.get_x_y(md_df)
            xs_train, xs_validation, ys_train, ys_validation = train_test_split(xs, ys, test_size=0.2, random_state=1)
            self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = xs_train, xs_validation, ys_train, ys_validation

        with self.session.as_default() as sess:
            real_ys = np.argmax(self.ys_validation, axis=1)

            # self.logger.info("批量预测2")
            # pred_ys = np.argmax(self.model.predict_label(self.xs_validation), axis=1) 与 evaluate 结果刚好相反
            # 因此不使用 predict_label 函数
            # pred_ys = np.argmax(self.model.predict(self.xs_validation), axis=1)
            # self.logger.info("模型训练基准日期：%s，validation accuracy: %.2f%%",
            #             self.trade_date_last_train, sum(pred_ys == real_ys) / len(pred_ys) * 100)
            # self.logger.info("pred/real: \n%s\n%s", pred_ys, real_ys)

            self.logger.info("独立样本预测(predict_latest)")
            pred_ys = []
            for idx, y in enumerate(self.ys_validation):
                x = self.xs_validation[idx:idx + 1, :, :]
                pred_y = self.model.predict(x)
                pred_ys.extend(np.argmax(pred_y, axis=1))

            pred_ys = np.array(pred_ys)
            self.logger.info("模型训练基准日期：%s，validation accuracy: %.2f%%",
                             self.trade_date_last_train, sum(pred_ys == real_ys) / len(pred_ys) * 100)
            # self.logger.info("pred: \n%s\n%s", pred_ys, real_ys)

    def predict_latest(self, factor_df):
        """
        计算最新一个 X，返回分类结果
        二分类，返回 0 未知 / 1 下跌 / 2 上涨
        三分类，返回 0 震荡 / 1 下跌 / 2 上涨
        :param factor_df:
        :return:
        """
        factors, _ = self.get_factor_array(factor_df, tail_n=self.n_step)
        x = self.get_batch_xs(factors)
        pred_y = np.argmax(self.model.predict(x), axis=1)[-1]
        # is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # return is_buy, is_sell
        return pred_y

    def on_prepare_min1(self, md_df, context):
        if md_df is None:
            return
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        indexed_df.index = pd.DatetimeIndex(indexed_df.index)
        self.load_train_test(indexed_df, enable_load_model=self.enable_load_model_if_exist)

    def load_train_test(self, indexed_df, enable_load_model, rebuild_model=False, enable_train_if_load_not_suss=True,
                        enable_train_even_load_succ=False, enable_test=False):
        if rebuild_model:
            self.get_model(rebuild_model=True)

        trade_date = str_2_date(indexed_df.index[-1])
        # 加载模型
        if enable_load_model:
            is_load = self.load_model_if_exist(trade_date)
        else:
            is_load = False

        if enable_train_even_load_succ or (enable_train_if_load_not_suss and not is_load):
            factor_df_dic = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                       trade_date_series=self.trade_date_series,
                                       delivery_date_series=self.delivery_date_series, do_multiple_factors=True)
            factor_df = factor_df_dic[1]
            num = 0
            while True:
                num += 1
                if num > 1:
                    self.get_model(rebuild_model=True)
                # 训练模型
                train_acc, val_acc = self.train(factor_df_dic, predict_test_random_state=num)
                if self.over_fitting_train_acc is not None and train_acc > self.over_fitting_train_acc:
                    self.logger.warning('第 %d 次训练，训练集精度 train_acc=%.2f%% 过高，可能存在过拟合，重新采样训练',
                                        num, train_acc * 100)
                    continue
                if self.validation_accuracy_base_line is not None:
                    if val_acc < self.validation_accuracy_base_line:
                        self.logger.warning('第 %d 次训练，训练结果不及预期，重新采样训练', num)
                        continue
                    # elif train_acc - val_acc > 0.15 and val_acc < 0.75:
                    #     self.logger.warning('第 %d 次训练，train_acc=%.2f%%, val_acc=%.2f%% 相差大于 15%% 且验证集正确率小于75%%，重新采样训练',
                    #                    num, train_acc * 100, val_acc * 100)
                    #     continue
                    else:
                        break
                else:
                    break

            self.save_model(trade_date)
            self.trade_date_last_train = trade_date
        else:
            factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                   trade_date_series=self.trade_date_series,
                                   delivery_date_series=self.delivery_date_series)
            train_acc, val_acc = self.valid_model_acc(factor_df)

        self.trade_date_acc_list[trade_date] = [train_acc, val_acc]

        # enable_test 默认为 False
        # self.valid_model_acc(factor_df) 以及完全取代 self.predict_test
        # self.predict_test 仅用于内部测试使用
        if enable_test:
            self.predict_test(factor_df)

        return factor_df

    def on_min1(self, md_df, context):
        if self.do_nothing_on_min_bar:  # 仅供调试使用
            return

        # 数据整理
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        indexed_df.index = pd.DatetimeIndex(indexed_df.index)
        # 获取最新交易日
        trade_date = str_2_date(indexed_df.index[-1])
        days_after_last_train = (trade_date - self.trade_date_last_train).days
        if self.retrain_period is not None and 0 < self.retrain_period < days_after_last_train:
            # 重新训练
            self.logger.info('当前日期 %s 距离上一次训练 %s 已经过去 %d 天，重新训练',
                             trade_date, self.trade_date_last_train, days_after_last_train)
            factor_df = self.load_train_test(indexed_df, rebuild_model=True,
                                             enable_load_model=self.enable_load_model_if_exist)
        else:
            factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                   trade_date_series=self.trade_date_series,
                                   delivery_date_series=self.delivery_date_series)

        # 预测
        pred_mark = self.predict_latest(factor_df)
        is_holding, is_buy, is_sell = pred_mark == 0, pred_mark == 1, pred_mark == 2
        # self.logger.info('%s is_buy=%s, is_sell=%s', trade_date, str(is_buy), str(is_sell))
        close = md_df['close'].iloc[-1]
        instrument_id = context[ContextKey.instrument_id_list][0]
        if is_buy:  # is_buy
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_target_position = False
            if no_target_position:
                self.open_long(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_sell:  # is_sell
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_holding:
            self.logger.debug("%s %s * * %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

    def on_min1_release(self, md_df):
        """
        增加模型对未来数据预测成功率走势图展示
        :param md_df:
        :return:
        """
        if md_df is None or md_df.shape[0] == 0:
            self.logger.warning('md_df is None or shape[0] == 0')
            return
        else:
            self.logger.debug('md_df.shape= %s', md_df.shape)

        # 获取各个模型训练时间点及路径
        date_file_path_pair_list = self.get_date_file_path_pair_list()
        if len(date_file_path_pair_list) > 0:
            # 按日期排序
            date_file_path_pair_list.sort(key=lambda x: x[0])

        # 建立数据集
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        trade_date_end = indexed_df.index[-1]
        factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                               trade_date_series=self.trade_date_series,
                               delivery_date_series=self.delivery_date_series)
        xs, ys_onehot, trade_date_index = self.get_x_y(factor_df)
        ys = np.argmax(ys_onehot, axis=1)
        data_len = len(trade_date_index)
        if data_len == 0:
            self.logger.warning('ys 长度为0，请检查是否存在数据错误')
            return
        trade_date2_list = [_[0] for _ in date_file_path_pair_list][1:]
        trade_date2_list.append(None)
        # 预测结果
        self.logger.info("按日期分段验证检验预测结果")
        pred_ys_tot, real_ys_tot, img_meta_dic_list = [], [], []
        # 根据模型 trade_date_last_train 进行分段预测，并将结果记录到 pred_ys
        for num, ((trade_date_last_train, file_path, predict_test_random_state), trade_date_next) in enumerate(zip(
                date_file_path_pair_list, trade_date2_list)):
            # 以模型训练日期为基准，后面的数据作为验证集数据（样本外数据）
            # 获取有效的日期范围 from - to
            range_from_arr = trade_date_index >= pd.to_datetime(trade_date_last_train)
            range_from_len = len(range_from_arr)
            if range_from_len == 0:  # range_from_len 应该与 trade_date_list_count 等长度，所以这个条件应该永远不会满足
                self.logger.error('总共%d条数据，%s 开始后面没有可验证数据', data_len, trade_date_last_train)
                continue
            true_count = sum(range_from_arr)
            self.logger.debug("len(range_from)=%d, True Count=%d", len(range_from_arr), true_count)
            if true_count == 0:
                self.logger.warning('总共%d条数据，%s 开始后面没有可验证数据', data_len, trade_date_last_train)
                continue
            # 自 trade_date_last_train 起的所有有效日期
            trade_date_list_sub = trade_date_index[range_from_arr]

            # 获取 in_range，作为 range_from, range_to 的交集
            if trade_date_next is None:
                in_range_arr = None
                in_range_count = true_count
            else:
                in_range_arr = trade_date_list_sub < pd.to_datetime(trade_date_next)
                in_range_count = sum(in_range_arr)
                if in_range_count == 0:
                    self.logger.warning('总共%d条数据，[%s - %s) 之间没有可用数据',
                                        data_len, trade_date_last_train, trade_date_next)
                    continue
                else:
                    self.logger.debug('总共%d条数据，[%s - %s) 之间有 %d 条数据将被验证 model path:%s',
                                      data_len, trade_date_last_train, trade_date_next, in_range_count, file_path)

            # 获取当前时段对应的 xs
            # 进行验证时，对 range_from 开始的全部数据进行预测，按照 range_to 为分界线分区着色显示
            xs_sub, real_ys = xs[range_from_arr, :, :], ys[range_from_arr]
            close_df = indexed_df.loc[trade_date_list_sub, 'close']

            # 加载模型
            is_load = self.load_model_if_exist(trade_date_last_train, enable_load_model_if_exist=True)
            if not is_load:
                self.logger.error('%s 模型加载失败：%s', trade_date_last_train, file_path)
                continue
            # 预测
            pred_ys_one_hot = self.model.predict(xs_sub)
            pred_ys = np.argmax(pred_ys_one_hot, axis=1)
            if in_range_arr is not None and in_range_count > 0:
                pred_ys_tot.extend(pred_ys[in_range_arr])
            else:
                pred_ys_tot.extend(pred_ys)

            # 为每一个时段单独验证成功率，以当前模型为基准，验证后面全部历史数据成功率走势
            if trade_date_next is None:
                split_point_list = None
            else:
                split_point_list = [close_df.index[0], trade_date_next, close_df.index[-1]]
            base_line_list = self.trade_date_acc_list[trade_date_last_train]
            img_file_path = show_dl_accuracy(real_ys, pred_ys, close_df, split_point_list,
                                             base_line_list=base_line_list)
            img_meta_dic_list.append({
                'img_file_path': img_file_path,
                'trade_date_last_train': trade_date_last_train,
                'module_file_path': file_path,
                'predict_test_random_state': predict_test_random_state,
                'split_point_list': split_point_list,
                'in_range_count': in_range_count,
                'trade_date_end': trade_date_end,
            })

        pred_ys_tot = np.array(pred_ys_tot)
        trade_date_last_train_first = pd.to_datetime(date_file_path_pair_list[0][0])
        split_point_list = [_[0] for _ in date_file_path_pair_list]
        split_point_list.append(trade_date_index[-1])
        # 获取 real_ys
        real_ys = ys[trade_date_index >= trade_date_last_train_first]
        close_df = indexed_df.loc[trade_date_index[trade_date_index >= trade_date_last_train_first], 'close']
        img_file_path = show_dl_accuracy(real_ys, pred_ys_tot, close_df, split_point_list)

        img_meta_dic_list.append({
            'img_file_path': img_file_path,
            'trade_date_last_train': trade_date_last_train_first,
            'module_file_path': date_file_path_pair_list[0][1],
            'predict_test_random_state': date_file_path_pair_list[0][2],
            'split_point_list': split_point_list,
            'in_range_count': close_df.shape[0],
            'trade_date_end': trade_date_end,
        })
        is_output_docx = True
        if is_output_docx:
            title = f"[{self.stg_run_id}] predict accuracy trend report"
            file_path = summary_release_2_docx(title, img_meta_dic_list)
            copy_file_to(file_path, self.base_folder_path)

    def get_date_file_path_pair_list(self):
        """
        目录结构：
        tf_saves_2019-06-05_16_21_39
          *   model_tfls
          *       *   2012-12-31
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *       *   2013-02-28
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *   tensorboard_logs
          *       *   2012-12-31_496[1]_20190605_184316
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
          *       *   2013-02-28_496[1]_20190605_184716
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
        :return:
        """
        # 获取全部文件名
        pattern = re.compile(r'model_[-]?\d+_\d+.tfl')
        date_file_path_pair_list, model_name_set = [], set()
        for folder_name in os.listdir(self.model_folder_path):
            folder_path = os.path.join(self.model_folder_path, folder_name)
            if os.path.isdir(folder_path):
                try:
                    # 获取 trade_date_last_train
                    key = str_2_date(folder_name)
                    for file_name in os.listdir(folder_path):
                        # 对下列有效文件名，匹配结果："model_-54_51.tfl"
                        # model_-54_51.tfl.data-00000-of-00001
                        # model_-54_51.tfl.index
                        # model_-54_51.tfl.meta
                        m = pattern.search(file_name)
                        if m is None:
                            continue
                        model_name = m.group()
                        if key in model_name_set:
                            continue
                        model_name_set.add(key)
                        # 获取 model folder_path
                        file_path = os.path.join(folder_path, model_name)
                        # 获取 predict_test_random_state
                        for log_folder_path in os.listdir(self.tensorboard_dir):
                            if log_folder_path.find(folder_name) == 0:
                                predict_test_random_state = int(log_folder_path.split('[')[1].split(']')[0])
                                break
                        else:
                            predict_test_random_state = None

                        date_file_path_pair_list.append([key, file_path, predict_test_random_state])
                except:
                    pass

        return date_file_path_pair_list



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-6-17 下午1:57
@File    : q_learn.py
@contact : mmmaaaggg@163.com
@desc    : Q Learning Table and Runner
"""

import logging
import os
from typing import Iterable

import numpy as np
import pandas as pd
from ibats_utils.mess import date_2_str, load_class
from ibats_common.example.rl_stg.v3 import module_version

logger = logging.getLogger(__name__)


def try_hash_or_str(v):
    try:
        if isinstance(v, dict):
            keys = list(v.keys())
            keys.sort()
            ret_str = "{" + ",".join([f"{try_hash_or_str(k)}:{try_hash_or_str(v[k])}" for k in keys]) + "}"
        elif isinstance(v, set):
            ret_str = "{" + ",".join([f"{try_hash_or_str(k)}" for k in v]) + "}"
        elif isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
            ret_str = v
        elif isinstance(v, Iterable):
            ret_str = "[" + ",".join([f"{try_hash_or_str(k)}" for k in v]) + "]"
        else:
            ret_str = hash(v)

        return ret_str
    except TypeError:
        return str(v)


def singleton(cls):
    instance = {}

    def _singleton(*args, **kw):
        keys = list(kw.keys())
        keys.sort()
        key = (cls, try_hash_or_str(args), try_hash_or_str(kw))
        if key not in instance:
            logger.debug("单例 %s", key)
            instance[key] = cls(*args, **kw)
        else:
            logger.debug("单例 已存在%s", key)
        return instance[key]

    return _singleton


@singleton
class QLearningTable:
    """
    Q-Learn 的实现类，完成获取state、reward、learn 等各种操作
    QLearningTable 采用的是带参数单利模式，
    因此，通过创建同样参数的实例，即可获取的在Env环境下训练后的 QLearningTable 结果
    """

    def __init__(self, actions, key=None, learning_rate=0.01, reward_decay=0.9, e_greedy=0.9):
        self.actions = actions  # a list
        self.key = key  # 对于每一个Q table 的标识同样参数的情况下，区别不同q table使用
        self.lr = learning_rate
        self.gamma = reward_decay
        self.epsilon = e_greedy
        self.q_table = pd.DataFrame(columns=self.actions, dtype=np.float64)

    def choose_action(self, observation):
        self.check_state_exist(observation)
        # action selection
        if np.random.uniform() < self.epsilon:
            # choose best action
            state_action = self.q_table.loc[observation, :]
            # some actions may have the same value, randomly choose on in these actions
            action = np.random.choice(state_action[state_action == np.max(state_action)].index)
        else:
            # choose random action
            action = np.random.choice(self.actions)
        return action

    def learn(self, s, a, r, s_):
        self.check_state_exist(s)
        self.check_state_exist(s_)
        q_predict = self.q_table.loc[s, a]
        if s_ != 'terminal':
            q_target = r + self.gamma * self.q_table.loc[s_, :].max()  # next state is not terminal
        else:
            q_target = r  # next state is terminal
        self.q_table.loc[s, a] += self.lr * (q_target - q_predict)  # update

    def check_state_exist(self, state):
        if state not in self.q_table.index:
            # append new state to q table
            self.q_table = self.q_table.append(
                pd.Series(
                    [0] * len(self.actions),
                    index=self.q_table.columns,
                    name=state,
                )
            )

    def _get_file_path(self, key):
        from ibats_common.backend.mess import get_folder_path
        folder_path = os.path.join(get_folder_path('example', create_if_not_found=False), 'module_data', module_version)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        file_name = f"q_table_{key}_{len(self.actions)}.csv"
        file_path = os.path.join(folder_path, file_name)
        return file_path

    def save(self):
        file_path = self._get_file_path(self.key)
        self.q_table.to_csv(file_path, index=False)

    def load(self):
        file_path = self._get_file_path(self.key)
        if os.path.exists(file_path):
            self.q_table = pd.read_csv(file_path)
            self.q_table.columns = self.q_table.columns.astype(np.int)
            return True
        else:
            return False

    def reset(self):
        self.q_table = pd.DataFrame(columns=self.actions, dtype=np.float64)


class Env:
    """
    Env 用于构建一个策略模拟环境，在该环境下利用 get_stg_handler 产生相应的 stg_handler 循环执行，
    以便对 QLearningTable 进行循环训练，
    QLearningTable 采用的是带参数单利模式，
    因此，通过创建同样参数的实例，即可获取的在Env环境下训练后的 QLearningTable 结果
    """

    def __init__(self, date_from, date_to, get_stg_handler, q_table_key=None):
        self.action_space = ['empty', 'hold_long', 'hold_short']
        self.n_actions = len(self.action_space)
        self.q_table_key = q_table_key
        self.stg_handler = None
        self.train_date_from, self.train_date_to = date_2_str(date_from), date_2_str(date_to)
        self.get_stg_handler = get_stg_handler
        self._build()

    def _build(self):
        # 初始化策略处理器
        get_stg_handler = self.get_stg_handler
        self.stg_handler = stg_handler = get_stg_handler(retrain_period=0, q_table_key=self.q_table_key)
        return stg_handler

    def set_stop(self):
        if self.stg_handler is not None:
            stg_run_id = self.stg_handler.stg_run_id
            if self.stg_handler.is_alive():
                logger.info('结束旧任务 stg_run_id=%d ...', stg_run_id)
                self.stg_handler.keep_running = False
                self.stg_handler.join()
                logger.info('结束旧任务 stg_run_id=%d 完成', stg_run_id)
            # 输出绩效报告
            # from ibats_common.analysis.summary import summary_stg_2_docx
            # from ibats_utils.mess import open_file_with_system_app
            # file_path = summary_stg_2_docx(stg_run_id, enable_clean_cache=False)
            # if file_path is not None:
            #     open_file_with_system_app(file_path)

    def reset_and_start(self):
        self.set_stop()
        stg_handler = self._build()
        stg_run_id = stg_handler.stg_run_id
        stg_handler.start()
        logging.info("执行开始 stg_run_id = %d", stg_run_id)

    def destroy(self):
        self.set_stop()


def _test_env(trade_date_from='2010-1-1', trade_date_to='2018-10-18'):
    from ibats_common.example.rl_stg.v3.rl_stg import get_stg_handler
    env = Env(trade_date_from, trade_date_to, get_stg_handler, q_table_key=trade_date_to)
    action_space = ['empty', 'hold_long', 'hold_short']
    actions = list(range(len(action_space)))
    q_learn_class = load_class(module_name=f'ibats_common.example.reinforcement_learning.{module_version}.q_learn',
                               class_name='QLearningTable')
    q_learn = q_learn_class(actions=actions, key=trade_date_to)
    for episode in range(2):
        # fresh env
        env.reset_and_start()
        # logger.info("q_learn.q_table\n%s", q_learn.q_table)
        print('q_learn.q_table')
        print(q_learn.q_table)

    # end of game
    env.destroy()
    logger.info('RL Over')


if __name__ == "__main__":
    _test_env()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/11/7 10:42
@File    : ma_cross_stg.py
@contact : mmmaaaggg@163.com
@desc    : V2主要是在V1基础上进行了结构重整，将 q_learn中的部分代码转移到 rl_stg 中
"""
from ibats_utils.mess import load_class, date_2_str
import numpy as np
from ibats_common.common import BacktestTradeMode, ContextKey, Direction, CalcMode
from ibats_common.example.rl_stg.v3.q_learn import QLearningTable, Env
from ibats_common.example.rl_stg.v3 import module_version
from ibats_common.strategy import StgBase
from ibats_common.strategy_handler import strategy_handler_factory
from ibats_local_trader.agent.md_agent import *
from ibats_local_trader.agent.td_agent import *

logger = logging.getLogger(__name__)


class RLStg(StgBase):

    def __init__(self, module_name, class_name, retrain_period=0, unit=1, q_table_key=None):
        """

        :param module_name:
        :param class_name:
        :param retrain_period: 0 不进行训练，int > 0 每隔 retrain_period 天重新训练
        :param unit:
        """
        super().__init__()
        self.unit = unit
        self.state = None
        self.module_name, self.class_name = module_name, class_name
        self.rl_handler_class = load_class(module_name, class_name)
        # RLHandler4Train 作为 RLHandler 的子类，完全支持 RLHandler 所有功能，这里将其区分开来
        # 主要是为了防止视觉上混淆，以及理解上的偏差
        if retrain_period == 0:
            # 无训练功能，仅进行买入卖出动作判断，供 Env 环境下训练时被调用
            self.rl_handler = self.rl_handler_class(q_table_key=q_table_key)
        else:
            # 有训练功能，将定期调用 RLHandler4Train 启动 Env 环境进行训练
            self.rl_handler = self.rl_handler_class(
                retrain_period=retrain_period, get_stg_handler=get_stg_handler, q_table_key=q_table_key)

    def on_prepare_min1(self, md_df, context):
        self.rl_handler.init_state(md_df)

    def long_position(self, close, instrument_id):
        position_date_pos_info_dic = self.get_position(instrument_id)
        no_holding_target_position = True
        if position_date_pos_info_dic is not None:
            for position_date, pos_info in position_date_pos_info_dic.items():
                if pos_info.position == 0:
                    continue
                direction = pos_info.direction
                if direction == Direction.Short:
                    self.close_short(instrument_id, close, pos_info.position)
                elif direction == Direction.Long:
                    no_holding_target_position = False
        if no_holding_target_position:
            self.open_long(instrument_id, close, self.unit)

    def short_position(self, close, instrument_id):
        position_date_pos_info_dic = self.get_position(instrument_id)
        no_holding_target_position = True
        if position_date_pos_info_dic is not None:
            for position_date, pos_info in position_date_pos_info_dic.items():
                if pos_info.position == 0:
                    continue
                direction = pos_info.direction
                if direction == Direction.Long:
                    self.close_long(instrument_id, close, pos_info.position)
                elif direction == Direction.Short:
                    no_holding_target_position = False
        if no_holding_target_position:
            self.open_short(instrument_id, close, self.unit)

    def empty_position(self, close, instrument_id):
        position_date_pos_info_dic = self.get_position(instrument_id)
        if position_date_pos_info_dic is not None:
            for position_date, pos_info in position_date_pos_info_dic.items():
                if pos_info.position == 0:
                    continue
                direction = pos_info.direction
                if direction == Direction.Long:
                    self.close_long(instrument_id, close, pos_info.position)
                elif direction == Direction.Short:
                    self.close_short(instrument_id, close, pos_info.position)

    def on_min1(self, md_df, context):
        close = md_df['close'].iloc[-1]
        action = self.rl_handler.choose_action(md_df)

        instrument_id = context[ContextKey.instrument_id_list][0]
        if action == 1:
            # 做多
            self.long_position(close, instrument_id)
        elif action == 2:
            # 做空
            self.short_position(close, instrument_id)
        elif action == 0:
            # 空仓
            self.empty_position(close, instrument_id)


def get_stg_handler(retrain_period, q_table_key=None):
    from ibats_common.backend.mess import get_folder_path
    import os
    # 参数设置
    instrument_type = 'RB'
    run_mode = RunMode.Backtest_FixPercent
    calc_mode = CalcMode.Normal
    if retrain_period == 0:
        strategy_params = {'unit': 1,
                           'module_name': f'ibats_common.example.rl_stg.{module_version}.rl_stg',
                           'class_name': 'RLHandler',
                           'q_table_key': q_table_key}
    else:
        strategy_params = {'unit': 1,
                           'module_name': f'ibats_common.example.rl_stg.{module_version}.rl_stg',
                           'class_name': 'RLHandler4Train',
                           'q_table_key': q_table_key,
                           'retrain_period': retrain_period
                           }

    md_agent_params_list = [{
        'md_period': PeriodType.Min1,
        'instrument_id_list': [instrument_type],
        'datetime_key': 'trade_date',
        'init_md_date_from': '1995-1-1',  # 行情初始化加载历史数据，供策略分析预加载使用
        'init_md_date_to': '2010-1-1',
        # 'C:\GitHub\IBATS_Common\ibats_common\example\ru_price2.csv'
        'file_path': os.path.abspath(os.path.join(
            get_folder_path('example', create_if_not_found=False), 'data', 'RB.csv')),
        'symbol_key': 'instrument_type',
    }]
    if run_mode == RunMode.Realtime:
        trade_agent_params = {
        }
        strategy_handler_param = {
        }
    elif run_mode == RunMode.Backtest:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            'init_cash': 1000000,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2010-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    else:
        # RunMode.Backtest_FixPercent
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2010-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }

    # 初始化策略处理器
    stghandler = strategy_handler_factory(
        stg_class=RLStg,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        exchange_name=ExchangeName.LocalFile,
        run_mode=run_mode,
        trade_agent_params=trade_agent_params,
        strategy_handler_param=strategy_handler_param,
    )
    return stghandler


class RLHandler:
    """
    RLHandler 是对 QLearningTable 的逻辑封装
    在策略使用是，可以方便的进行初始化及调用（将初始化，训练、获取 state、reward 等操作进行的同一封装）
    从而简化了 RLStg 策略中的代码复杂度
    """

    def __init__(self, q_table_key=None):
        """
        :param q_table_key:
        """
        action_space = ['empty', 'hold_long', 'hold_short']
        self.actions = list(range(len(action_space)))
        self.q_table_key = q_table_key
        self.ql_table = None
        self.last_state = None
        self.last_action = None
        self.enable_load_if_exist = True

    def init_state(self, md_df: pd.DataFrame):
        # 每一次新 episode 需要重置 state, action
        self.last_state = None
        self.last_action = None

        _ = self.choose_action(md_df)

    def init_ql_table(self, trade_date_to=None):
        key = self.q_table_key if self.q_table_key is not None else date_2_str(trade_date_to)
        self.ql_table = QLearningTable(actions=self.actions, key=key)

    def get_state_reward(self, md_df: pd.DataFrame):
        close_s = md_df['close']
        close_log_arr = np.log(close_s.iloc[-6:])
        close_latest = close_log_arr.iloc[-1]
        state = int((close_latest - close_log_arr.iloc[0]) * 1000)
        if self.last_action is not None:
            if self.last_action == 0:
                reward = 0
            elif self.last_action == 1:
                reward = int((close_latest - close_log_arr.iloc[-2]) * 1000)
            elif self.last_action == 2:
                reward = - int((close_latest - close_log_arr.iloc[-2]) * 1000)
            else:
                raise ValueError(f'last_action = {self.last_action} 值不合法')
        else:
            reward = 0

        return state, reward

    def choose_action(self, md_df: pd.DataFrame):
        trade_date_to = pd.to_datetime(md_df['trade_date'].iloc[-1])
        if self.ql_table is None:
            self.init_ql_table(trade_date_to)

        state, reward = self.get_state_reward(md_df)
        if self.last_state is not None:
            self.ql_table.learn(self.last_state, self.last_action, reward, state)

        self.last_action = self.ql_table.choose_action(state)
        self.last_state = state
        return self.last_action


class RLHandler4Train(RLHandler):
    """
    RLHandler 是对 QLearningTable 的逻辑封装
    在策略使用是，可以方便的进行初始化及调用（将初始化，训练、获取 state、reward 等操作进行的同一封装）
    从而简化了 RLStg 策略中的代码复杂度
    """

    def __init__(self, retrain_period, get_stg_handler, episode_count=3, q_table_key=None):
        """

        :param retrain_period: 0 不进行训练，int > 0 每隔 retrain_period 天重新训练
        :param episode_count:
        :param get_stg_handler: 用于构建 stg_handler 的函数，仅当 retrain_period > 0 时才会有用
        :param q_table_key:
        """
        RLHandler.__init__(self, q_table_key=q_table_key)
        self.retrain_period = pd.Timedelta(days=retrain_period)
        if get_stg_handler is None:
            raise ValueError('get_stg_handler must be not None')
        self.get_stg_handler = get_stg_handler
        self.do_train = retrain_period > 0 and get_stg_handler is not None
        self.train_date_latest = None
        self.train_date_from = None
        self.episode_count = episode_count

    def init_state(self, md_df: pd.DataFrame):
        # 每一次新 episode 需要重置 state, action
        self.last_state = None
        self.last_action = None
        if self.do_train:
            # for_train == False 当期为策略运行使用，在 on_prepare 阶段以及 on_period 定期进行重新训练
            trade_date_s = md_df['trade_date']
            self.train_date_from = pd.to_datetime(trade_date_s.iloc[0]) + self.retrain_period
            trade_date_to = trade_date_s.iloc[-1]
            self.train(date_2_str(trade_date_to))

        _ = self.choose_action(md_df)

    def train(self, trade_date_to):
        """
        具体功能参见 ibats_common.example.rl_stg.q_learn import main
        :param trade_date_to:
        :return:
        """
        if self.enable_load_if_exist:
            self.init_ql_table(trade_date_to)
            is_loaded = self.ql_table.load()
        else:
            is_loaded = False

        if self.ql_table is None:
            self.init_ql_table(trade_date_to)

        if not is_loaded:
            logger.info('开始训练：[%s, %s]', self.train_date_from, trade_date_to)
            env = Env(self.train_date_from, trade_date_to, self.get_stg_handler, q_table_key=self.ql_table.key)
            for episode in range(self.episode_count):
                # fresh env
                env.reset_and_start()
                # logger.info("\n%s", self.q_table.q_table)

            # end of game
            env.destroy()
            logger.info('RL Over')
            self.ql_table.save()
        # 设置最新训练日期
        self.train_date_latest = pd.to_datetime(trade_date_to)

    def choose_action(self, md_df: pd.DataFrame):
        trade_date_to = pd.to_datetime(md_df['trade_date'].iloc[-1])
        if self.do_train and trade_date_to > (self.train_date_latest + self.retrain_period):
            # for_train == False 当期为策略运行使用，在 on_prepare 阶段以及 on_period 定期进行重新训练
            self.train(date_2_str(trade_date_to))
        elif self.ql_table is None:
            self.init_ql_table(trade_date_to)

        state, reward = self.get_state_reward(md_df)
        if self.last_state is not None:
            self.ql_table.learn(self.last_state, self.last_action, reward, state)

        self.last_action = self.ql_table.choose_action(state)
        self.last_state = state
        return self.last_action


def _test_rl_handler(trade_date_from='2010-1-1', trade_date_to='2018-10-18'):
    from ibats_common.example.data import load_data
    trade_date_from_ = str_2_date(trade_date_from)
    trade_date_to_ = str_2_date(trade_date_to)
    df = load_data('RB.csv')
    md_df = df[df['trade_date'].apply(lambda x: str_2_date(x) <= trade_date_to_)]
    trade_date_s = md_df['trade_date'].apply(lambda x: str_2_date(x))
    rl_handler = RLHandler(q_table_key=trade_date_to)
    rl_handler.init_state(md_df)
    trade_date_action_dic = {}
    for trade_date in trade_date_s[trade_date_s >= trade_date_from_]:
        md_df_curr = md_df[trade_date_s.apply(lambda x: x <= trade_date)]
        action = rl_handler.choose_action(md_df_curr)
        trade_date_action_dic[trade_date] = action

    action_df = pd.DataFrame([trade_date_action_dic]).T
    print("action_df = \n", action_df)
    ql_table = QLearningTable(actions=rl_handler.actions, key=trade_date_to)
    # assert ql_table.q_table.shape[0] == 0
    # ql_table_class = load_class(module_name='ibats_common.example.rl_stg.q_learn',
    #                            class_name='QLearningTable')
    # ql_table = ql_table_class(actions=rl_handler.actions, key=trade_date_to)
    assert ql_table.q_table.shape[0] > 0
    print("ql_table.q_table.shape=", ql_table.q_table.shape)
    assert ql_table.q_table.shape == rl_handler.ql_table.q_table.shape
    # print(ql_table.q_table)


def _test_rl_handler_4_train(trade_date_from='2010-1-1', trade_date_to='2018-10-18'):
    from ibats_common.example.rl_stg.v1.rl_stg import get_stg_handler
    from ibats_common.example.data import load_data
    trade_date_from_ = str_2_date(trade_date_from)
    trade_date_to_ = str_2_date(trade_date_to)
    df = load_data('RB.csv')
    md_df = df[df['trade_date'].apply(lambda x: str_2_date(x) <= trade_date_to_)]
    trade_date_s = md_df['trade_date'].apply(lambda x: str_2_date(x))
    rl_handler = RLHandler4Train(retrain_period=360, get_stg_handler=get_stg_handler, q_table_key=trade_date_to)
    rl_handler.init_state(md_df)
    trade_date_action_dic = {}
    for trade_date in trade_date_s[trade_date_s >= trade_date_from_]:
        md_df_curr = md_df[trade_date_s.apply(lambda x: x <= trade_date)]
        action = rl_handler.choose_action(md_df_curr)
        trade_date_action_dic[trade_date] = action

    action_df = pd.DataFrame([trade_date_action_dic]).T
    print("action_df = \n", action_df)
    # ql_table = QLearningTable(rl_handler.actions, key=trade_date_to)
    ql_table_class = load_class(module_name=f'ibats_common.example.rl_stg.{module_version}.q_learn',
                                class_name='QLearningTable')
    ql_table = ql_table_class(actions=rl_handler.actions, key=trade_date_to)
    assert ql_table.q_table.shape[0] > 0
    print("ql_table.q_table.shape=", ql_table.q_table.shape)
    # rl_handler.init_ql_table()
    assert ql_table.q_table.shape == rl_handler.ql_table.q_table.shape
    # print(ql_table.q_table)


def _test_use(is_plot):
    stghandler = get_stg_handler(retrain_period=7)
    stghandler.start()
    time.sleep(10)
    stghandler.keep_running = False
    stghandler.join()
    stg_run_id = stghandler.stg_run_id
    logging.info("执行结束 stg_run_id = %d", stg_run_id)

    if is_plot:
        # from ibats_common.analysis.plot_db import show_order, show_cash_and_margin, show_rr_with_md
        # from ibats_common.analysis.summary import summary_rr
        # show_order(stg_run_id)
        # df = show_cash_and_margin(stg_run_id)
        # sum_df, symbol_rr_dic, save_file_path_dic = show_rr_with_md(stg_run_id)
        # for symbol, rr_df in symbol_rr_dic.items():
        #     col_transfer_dic = {'return': rr_df.columns}
        #     summary_rr(rr_df, figure_4_each_col=True, col_transfer_dic=col_transfer_dic)
        from ibats_common.analysis.summary import summary_stg_2_docx
        from ibats_utils.mess import open_file_with_system_app
        file_path = summary_stg_2_docx(stg_run_id)
        if file_path is not None:
            open_file_with_system_app(file_path)

    return stg_run_id


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    _test_use(is_plot=True)
    # _test_rl_handler()
    # _test_rl_handler_4_train()
